"""
NanoStreamQueue module
=====================

These are queues that form the directed edges between nodes.
"""

import queue
import uuid
import logging
import time
from nanostream.message.message import NanoStreamMessage
from nanostream.message.batch import BatchStart, BatchEnd
from nanostream import node


QUEUE_TIME_WINDOW = 100


class NanoStreamQueue:
    """
    """

    def __init__(self, max_queue_size, name=None):
        self.queue = queue.Queue(max_queue_size)
        self.name = name or uuid.uuid4().hex
        self.source_node = None
        self.target_node = None
        self.queue_times = []  # Time messages spend in queue


    @property
    def empty(self):
        return self.queue.empty()

    def get(self):
        try:
            message = self.queue.get(block=False)
            self.queue_times.append(time.time() - message.time_queued)
            self.queue_times = self.queue_times[-1 * QUEUE_TIME_WINDOW:]
        except queue.Empty:
            message = None
        logging.debug('Retrieved message: ' + str(message))
        return message

    def put(self, message, *args, previous_message=None, **kwargs):
        '''
        Places a message on the output queues. If the message is ``None``,
        then the queue is skipped.

        Messages are ``NanoStreamMessage`` objects; the payload of the
        message is message.message_content.
        '''
        if isinstance(message, (node.NothingToSeeHere,)):
            return
        elif previous_message is not None:
            previous_message = previous_message.message_content
        else:
            previous_message = {}
        if self.source_node.output_key is not None:
            message = {self.source_node.output_key: message}

        # Check if we need to retain the previous message in the keys of
        # this message, assuming we have dictionaries, etc.
        if self.source_node.retain_input:
            for key, value in previous_message.items():
                if key not in message:
                    message[key] = value
                elif (key in message and value is not None
                        and self.source_node.prefer_existing_value):
                    message[key] = value
                else:
                    pass


        message_obj = NanoStreamMessage(message)
        message_obj.time_queued = time.time()
        self.queue.put(message_obj)



'''
Network nodes module
====================

Classes that deal with sending and receiving data across the interwebs.
'''

import requests
import json
import logging
from requests.packages.urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

from nanostream.node import NanoNode
from nanostream.utils.helpers import SafeMap

additional_data_test = bool


class PaginatedHttpGetRequest:
    '''
    For handling requests in a semi-general way that require paging through
    lists of results and repeatedly making GET requests.
    '''

    def __init__(
        self, endpoint_template=None, additional_data_key=None,
        pagination_key=None, pagination_get_request_key=None,
            default_offset_value='', additional_data_test=bool):
        '''
        :ivar endpoint_template: (str) Template for endpoint URL, suitable
            for calling ``endpoint_template.format(**kwargs)``.
        :ivar additional_data_key: Key in JSON payload whose value
            indicates whether there are additional pages to request.
        :ivar pagination_key: Key in JSON payload where the current page
            (or next page) is indicated.
        :ivar pagination_get_request_key: Variable in URL GET request where
            we pass the offset for the next page.
        :ivar default_offset_value: Offset value for the first request.
            Usually this will be an empty string.
        :ivar additional_data_test: Function that is passed the value of
            ``additional_data_key``. It should return ``True`` if there are
            additional pages to request, ``False`` otherwise.
        '''
        self.endpoint_template = endpoint_template
        self.additional_data_key = additional_data_key
        self.pagination_key = pagination_key
        self.pagination_get_request_key = pagination_get_request_key
        self.default_offset_value = default_offset_value
        self.additional_data_test = additional_data_test

    def responses(self):
        '''
        Generator. Yields each response until empty.
        '''
        offset_set = set()
        session = requests.Session()
        retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[ 500, 502, 503, 504 ])

        session.mount('http://', HTTPAdapter(max_retries=retries))

        get_request_parameters = {
            self.pagination_get_request_key: self.default_offset_value}
        endpoint_url = self.endpoint_template.format(
            **get_request_parameters)
        logging.info('paginator url: ' + endpoint_url)
        out = session.get(endpoint_url)
        # out = out.json()
        out = json.loads(out.text)
        offset = out.get(self.pagination_key, None)
        offset_set.add(offset)

        while self.additional_data_key in out and additional_data_test(out[self.additional_data_key]):
            yield out
            try:
                offset = out[self.pagination_key]
                offset_set.add(offset)
            except KeyError:
                logging.debug('No offset key. Assuming this is normal.')
                break
            get_request_parameters = {
                self.pagination_get_request_key: offset}
            endpoint_url = self.endpoint_template.format(
                **get_request_parameters)
            logging.debug('paginator url: ' + endpoint_url)
            try:
                response = session.get(endpoint_url)
                out = response.json()
            except:
                logging.warning('Error parsing. Assuming this is the end of the responses.')
                break


class HttpGetRequest(NanoNode):
    '''
    Node class for making simple GET requests.
    '''

    def __init__(
        self,
        endpoint_template=None,
        endpoint_dict=None,
        json=True,
            **kwargs):
        self.endpoint_template = endpoint_template
        self.endpoint_dict = endpoint_dict or {}
        self.json = json
        self.endpoint_dict.update(self.endpoint_dict)

        super(HttpGetRequest, self).__init__(**kwargs)

    def process_item(self):
        '''
        The input to this function will be a dictionary-like object with
        parameters to be substituted into the endpoint string and a
        dictionary with keys and values to be passed in the GET request.

        Three use-cases:
        1. Endpoint and parameters set initially and never changed.
        2. Endpoint and parameters set once at runtime
        3. Endpoint and parameters set by upstream messages
        '''

        # Hit the parameterized endpoint and yield back the results
        formatted_endpoint = self.endpoint_template.format_map(SafeMap(**(self.message or {})))
        try:
            formatted_endpoint = formatted_endpoint.format_map(SafeMap(**(self.endpoint_dict or {})))
        except Exception as err:
            print('formatted endpoint: ' + formatted_endpoint)
            raise Exception()
        logging.info('Http GET request: {endpoint}'.format(endpoint=formatted_endpoint))
        get_response = requests.get(formatted_endpoint)
        try:
            output = get_response.json()
        except JSONDecodeError:
            output = get_response.text
        logging.info(formatted_endpoint + ' GET RESPONSE: ' + str(output) + str(type(output)))
        yield output


class HttpGetRequestPaginator(NanoNode):
    '''
    Node class for HTTP API requests that require paging through sets of
    results.
    '''

    def __init__(
        self,
        endpoint_dict=None,
        json=True,
        pagination_get_request_key=None,
        endpoint_template=None,
        additional_data_key=None,
        pagination_key=None,
        pagination_template_key=None,
        default_offset_value='',
            **kwargs):
        self.pagination_get_request_key = pagination_get_request_key
        self.additional_data_key = additional_data_key
        self.pagination_key = pagination_key
        self.endpoint_dict = endpoint_dict or {}
        self.endpoint_template = endpoint_template or ''
        self.default_offset_value = default_offset_value

        self.endpoint_template = self.endpoint_template.format_map(
            SafeMap(**self.endpoint_dict))

        super(HttpGetRequestPaginator, self).__init__(**kwargs)

    def process_item(self):
        self.requestor = PaginatedHttpGetRequest(
            pagination_get_request_key=self.pagination_get_request_key,
            endpoint_template=self.endpoint_template.format_map(
                SafeMap(**(self.message or {}))),
            additional_data_key=self.additional_data_key,
            pagination_key=self.pagination_key,
            default_offset_value=self.default_offset_value)

        for i in self.requestor.responses():
            logging.debug('paginator:' + str(self.name) + ' ' + str(i))
            if self.finished:
                break
            yield i



"""
Civis-specific node types
=========================

This is where any classes specific to the Civis API live.
"""

import logging
import os
import tempfile
import time
import csv
import sys
import uuid
import civis
from civis.base import DONE

from nanostream.node import *
from nanostream.node_classes.network_nodes import HttpGetRequestPaginator
from nanostream.utils.helpers import remap_dictionary, SafeMap, list_to_dict
from timed_dict.timed_dict import TimedDict

MONITOR_FUTURES_SLEEP = 5

class SendToCivis(NanoNode):
    def __init__(
        self,
        *args,
        civis_api_key=None,
        civis_api_key_env_var='CIVIS_API_KEY',
        database=None,
        schema=None,
        existing_table_rows='append',
        include_columns=None,
        dummy_run=False,
        block=False,
        max_errors=0,
        table=None,
        columns=None,
        remap=None,
        recorded_tables=TimedDict(timeout=30),
            **kwargs):
        self.civis_api_key = civis_api_key or os.environ[civis_api_key_env_var]
        self.include_columns = include_columns
        self.table = table
        self.dummy_run = dummy_run
        self.schema = schema
        self.max_errors = int(max_errors)
        self.existing_table_rows = existing_table_rows
        self.database = database
        self.block = block
        self.remap = remap
        self.recorded_tables = recorded_tables
        self.full_table_name = '.'.join([self.schema, self.table])
        self.columns = columns

        if self.civis_api_key is None and len(self.civis_api_key) == 0:
            raise Exception('Could not get a Civis API key.')

        super(SendToCivis, self).__init__(**kwargs)
        self.monitor_futures_thread = threading.Thread(
            target=SendToCivis.monitor_futures, args=(self, ), daemon=True)
        self.monitor_futures_thread.start()

    def setup(self):
        '''
        Not sure if we'll need this. We could get a client and pass it around.
        '''
        pass

    def monitor_futures(self):

        class DummyResult:
            def __init__(self):
                self.state = 'done'

        run = True
        while run:
            logging.debug('Checking future objects...')

            table_lock = threading.Lock()
            table_lock.acquire(blocking=True)

            try:
                table_items = list(self.recorded_tables.items())
            except RuntimeError:
                logging.warning('Runtime error in dictionary comprehension. Continuing.')
                continue

            for table_id, future_dict in table_items:
                future_obj = future_dict['future']
                row_list = future_dict['row_list']
                # logging.debug(future_obj.done())
                logging.debug('poller result:' + str(future_obj._state) + str(type(future_obj._state)))
                if future_obj._state != 'RUNNING':
                    if future_obj.failed():
                        logging.info('failer in SendToCivis: ' + str(future_obj.exception()))
                        self.status = 'error'  # Needs to be caught by Node class
                        run = False
            table_lock.release()
            time.sleep(MONITOR_FUTURES_SLEEP)
        for table_id, future_dict in list(self.recorded_tables.items()):
            setattr(future_dict['future'], 'done', lambda: True)
            future_dict['future'].set_result(DummyResult())
            # future_dict['future'].cleanup()
            # CivisFuture object, not ``Future`` -- this is why the problems!

    def process_item(self):
        '''
        Accept a bunch of dictionaries mapping column names to values.
        '''

        # with tempfile.NamedTemporaryFile(mode='w') as tmp:
        row_list = []
        if self.name == 'send_email_to_redshift':
            logging.info('send_email_to_redshift')
            logging.info(str(self.__message__))
        if len(self.__message__) == 0:
            yield self.message

        else:
            with tempfile.NamedTemporaryFile(mode='w') as tmp:
                if self.include_columns is not None:
                    fieldnames = self.include_columns
                elif self.columns is not None:
                    fieldnames = self.columns
                else:
                    try:
                        fieldnames = sorted(list(self.__message__[0].keys()))
                    except:
                        import pdb; pdb.set_trace()
                writer = csv.DictWriter(
                    tmp,
                    fieldnames=fieldnames,
                    extrasaction='ignore',
                    quoting=csv.QUOTE_ALL)
                writer.writeheader()
                row_list.append(fieldnames)
                for row in self.__message__:
                    # Optionally remap row here
                    if self.remap is not None:
                        row = remap_dictionary(row, self.remap)
                    #if 'is_contact' in row:  # Boom
                    #    row['is_contact'] = 'foobar'
                    writer.writerow(row)
                    row_list.append(row)  # Will this get too slow?
                tmp.flush()
                logging.info('to redshift: ' + str(self.name))
                logging.info(str(row_list))
                if not self.dummy_run:
                    fut = civis.io.csv_to_civis(
                        tmp.name,
                        self.database,
                        self.full_table_name,
                        max_errors=self.max_errors,
                        headers=True,
                        existing_table_rows=self.existing_table_rows)
                    table_id = uuid.uuid4()
                    self.recorded_tables[table_id.hex] = {
                        'row_list': row_list,
                        'future': fut}
                    if self.block:
                        while not fut.done():
                            time.sleep(1)
                else:
                    logging.info('Not sending to Redshift due to `dummy run`')
            yield self.message


class EnsureCivisRedshiftTableExists(NanoNode):

    def __init__(
        self,
        on_failure='exit',
        table=None,
        schema=None,
        columns=None,
        block=True,
            **kwargs):

        self.on_failure = on_failure
        self.table = table
        self.schema = schema
        self.columns = columns
        self.block = block
        if any(i is None for i in [
                on_failure, table, schema, columns]):
            raise Exception('Missing parameters.')
        super(EnsureCivisRedshiftTableExists, self).__init__(**kwargs)

    def process_item(self):
        for i in self.generator():
            yield i

    def generator(self):
        columns_spec = ', '.join(
            ['"{column_name}" {column_type} NULL'.format(
                column_name=column['column_name'],
                column_type=column['column_type']) for column in self.columns])
        create_statement = (
            '''CREATE TABLE IF NOT EXISTS "{schema}"."{table}" '''
            '''({columns_spec});'''.format(
                schema=self.schema,
                table=self.table,
                columns_spec=columns_spec))
        logging.debug('Ensuring table exists -- ' + create_statement)
        fut = civis.io.query_civis(create_statement, 'Greenpeace')
        result = fut.result()

        yield columns_spec


class FindValueInRedshiftColumn(NanoNode):

    def __init__(
        self,
        on_failure='exit',
        table=None,
        database=None,
        schema=None,
        column=None,
        choice='max',
            **kwargs):

        self.on_failure = on_failure
        self.table = table
        self.schema = schema
        self.database = database
        self.column = column
        self.database = database
        self.choice = choice.upper()

        if self.choice not in ['MAX', 'MIN']:
            raise Exception(
                'The `choice` parameter must be one of [MAX, MIN].')
        super(FindValueInRedshiftColumn, self).__init__(**kwargs)

    def process_item(self):
        for i in self.generator():
            yield i

    def generator(self):
        create_statement = (
            '''SELECT {choice}({column}) FROM {schema}.{table};'''.format(
                schema=self.schema,
                table=self.table,
                column=self.column,
                choice=self.choice.upper()))
        fut = civis.io.query_civis(create_statement, self.database)
        result = fut.result()
        value = (
            result['result_rows'][0][0]
            if len(result['result_rows']) > 0 and len(result['result_rows'][0]) > 0 else None)
        logging.info('FindValueInRedshiftColumn: ' + str(value))
        yield value


class CivisSQLExecute(NanoNode):
    '''
    Execute a SQL statement and return the results.
    '''

    def __init__(
        self,
        *args,
        sql=None,
        civis_api_key=None,
        civis_api_key_env_var='CIVIS_API_KEY',
        database=None,
        dummy_run=False,
        query_dict=None,
        returned_columns=None,
            **kwargs):
        self.sql = sql
        self.query_dict = query_dict or {}
        self.civis_api_key = civis_api_key or os.environ[civis_api_key_env_var]
        self.dummy_run = dummy_run
        self.database = database
        self.returned_columns = returned_columns

        if self.civis_api_key is None and len(self.civis_api_key) == 0:
            raise Exception('Could not get a Civis API key.')

        super(CivisSQLExecute, self).__init__(**kwargs)

    def process_item(self):
        '''
        Execute a SQL statement and return the result.
        '''
        sql_query = self.sql.format_map(SafeMap(**self.query_dict))
        sql_query = sql_query.format_map(SafeMap(**(self.message or {})))
        logging.info(sql_query)
        if not self.dummy_run:
            fut = civis.io.query_civis(
                sql_query,
                self.database)
            result = fut.result()
        else:
            logging.info('Not querying Redshift due to `dummy run`')
            result = None
        result_rows = result['result_rows']
        if self.returned_columns is not None:
            result_rows = [
                list_to_dict(row, self.returned_columns) for row in result_rows]
        else:
            result_rows = result['result_rows']
        yield {'result_rows': result_rows}


class CivisToCSV(NanoNode):
    '''
    Execute a SQL statement and return the results via a CSV file.
    '''

    def __init__(
        self,
        *args,
        sql=None,
        civis_api_key=None,
        civis_api_key_env_var='CIVIS_API_KEY',
        database=None,
        dummy_run=False,
        query_dict=None,
        returned_columns=None,
        include_headers=True,
        delimiter=',',
            **kwargs):
        self.sql = sql
        self.query_dict = query_dict or {}
        self.civis_api_key = civis_api_key or os.environ[civis_api_key_env_var]
        self.dummy_run = dummy_run
        self.database = database
        self.returned_columns = returned_columns
        self.include_headers = include_headers
        self.delimiter = delimiter

        if self.civis_api_key is None and len(self.civis_api_key) == 0:
            raise Exception('Could not get a Civis API key.')

        super(CivisToCSV, self).__init__(**kwargs)

    def process_item(self):
        '''
        Execute a SQL statement and return the result.
        '''
        sql_query = self.sql.format_map(SafeMap(**self.query_dict))
        sql_query = sql_query.format_map(SafeMap(**(self.message or {})))
        tmp_filename = uuid.uuid4().hex + '_tmp.csv'
        fut = civis.io.civis_to_csv(tmp_filename, sql_query, self.database)
        fut.result()
        #while fut._result == 'RUNNING':
        #    time.sleep(1)
        #logging.info('future state: ' + str(fut._state))
        try:
            csv_file = open(tmp_filename, 'r')
            csv_reader = csv.DictReader(csv_file)
            for row in csv_reader:
                yield row
            os.remove(tmp_filename)
        except FileNotFoundError:
            logging.info('FileNotFoundError in CivisToCSV')
            yield NothingToSeeHere()


if __name__ == '__main__':
    # Test so that we can get a better view into errors
    api_key = os.environ['CIVIS_API_KEY']
    fut = civis.io.csv_to_civis(
        'email_test_data.csv',
        'Greenpeace',
        'staging.email_raw',
        max_errors=0,
        headers=True,
        existing_table_rows='append')



'''
Decorator that eliminates the need to assign attributes to the keys
and values of kwargs in an ``__init__`` function.
'''

import inspect


class set_kwarg_attributes:
    '''
    Function decorator to be used on ``__init__`` methods. It examines the
    signature of the ``__init__`` method and any kwargs passed to the
    constructor. It creates attributes named after the keys, and assigns
    the corresponding values to those attributes.

    For example:

    ::

        class Foo:
            @set_kwarg_attributes()
            def __init__(self, foo='bar', bar='baz'):
                pass

    Calling ``foo = Foo()``, you'd find that ``foo.foo == 'bar`` and
    ``foo.bar == 'baz'``, without having to go through the tedious torture
    of explicitly writing boilerplate like ``self.foo = foo``.

    You can also exclude some kwargs from this magic by passing a list
    ``exclude=[...]`` to the decorator.
    '''

    def __init__(self, exclude=None):
        '''
        Constructor for the decorator.

        Args:
            exclude (list of str): Names of kwargs that will **not** be
                assigned to attributes of the object.
        '''
        self.exclude = exclude or []

    def __call__(self, f):
        def inner_function(_self, *args, **kwargs):
            updated_kwargs = {}
            for name, value in inspect.signature(f).parameters.items():
                if not isinstance(value.default, type):
                    updated_kwargs[name] = value.default
            updated_kwargs.update(kwargs)
            for kwarg, value in updated_kwargs.items():
                if kwarg in self.exclude:
                    continue
                setattr(_self, kwarg, value)
            f(_self, *args, **kwargs)

        return inner_function



'''
Helper module
*************

Misc. helper functions for other classes.
'''

import copy
import types
import time
import logging
import datetime
import pickle
import hashlib
import uuid


UNIX_EPOCH = datetime.datetime(month=1, year=1970, day=1)


def list_to_dict(some_list, list_of_keys):
    if len(some_list) != len(list_of_keys):
        raise Exception('Length of list elements and key list must be equal.')
    out = {
        list_of_keys[index]: item for index, item in enumerate(some_list)}
    return out


def timestamp_to_redshift(timestamp):
    if isinstance(timestamp, (str,)):
        return timestamp
    return timestamp.strftime('%b %d,%Y  %H:%M:%S')


def milliseconds_epoch_to_datetime(milliseconds_epoch):
    if isinstance(milliseconds_epoch, (datetime.datetime, str,)):
        return milliseconds_epoch
    logging.info('milliseconds_epoch_to_datetime: ' + str(milliseconds_epoch))
    out = UNIX_EPOCH + datetime.timedelta(seconds=(
        int(milliseconds_epoch)/1000))
    logging.info('milliseconds_epoch_to_datetime output: ' + str(out))
    return out


def seconds_epoch_to_datetime(seconds_epoch):
    return UNIX_EPOCH + datetime.timedelta(seconds=(seconds))


def to_bool(thing):
    if isinstance(thing, (str,)):
        return len(thing) > 0 and thing[0].lower() in ['t', 'y']
    elif isinstance(thing, (int, float,)):
        return thing > 0
    elif isinstance(thing, (bool,)):
        return thing
    else:
        raise Exception(
            'Do not know how to convert {thing} to bool'.format(
                thing=str(thing)))


def get_value(
    dictionary, path, delimiter='.',
        use_default_value=False, default_value=None):

    # dictionary = copy.deepcopy(dictionary)
    if isinstance(path, (str,)):
        path = path.split(delimiter)
    elif isinstance(path, (list, tuple,)):
        pass
    else:
        raise Exception('what?')
    if len(path) == 0:
        dictionary = dictionary
    else:
        for step in path:
            if dictionary is None or isinstance(dictionary, (dict,)):
                dictionary = (dictionary or {}).get(step, default_value)
            else:
                dictionary = dictionary[step.index]
    return dictionary


def hi(*args, **kwargs):
    return 'hi'


def pdb(*args, **kwargs):
    import pdb; pdb.set_trace()


def convert_date_format(date_string, source_format=None, target_format=None):
    return datetime.datetime.strptime(
        date_string, source_format).strftime(target_format)


def engaging_networks_date(result, **kwargs):
    '''
    This is a very redundant function, but it's here mostly to test the
    functionality of the ``post_process_function``.
    '''
    out = datetime.datetime.strptime(result[0]['date_modified'], '%Y-%m-%d')
    out = out.strftime('%Y%m%d')
    return out


def set_value(dictionary, path, value):
    for step in path[:-1]:
        if not isinstance(step, (ListIndex,)):
            if step not in dictionary:
                raise Exception('Path not found.')
            dictionary = dictionary[step]
        else:
            dictionary = dictionary[step.index]
    dictionary[
        path[-1] if not isinstance(path[-1], (ListIndex,))
        else path[-1].index] = value


def iterate_leaves(dictionary, keypath=None):
    keypath = keypath or []
    for key, value in dictionary.items():
        if not isinstance(value, (dict,)):
            yield keypath + [key], value
        else:
            for i in iterate_leaves(value, keypath=keypath + [key]):
                yield i


def remap_dictionary(
    source_dictionary, target_dictionary,
        use_default_value=False, default_value=None):
    target_dictionary = copy.deepcopy(target_dictionary)
    for path, value in iterate_leaves(target_dictionary):
        set_value(
            target_dictionary,
            path,
            get_value(
                source_dictionary,
                value,
                use_default_value=use_default_value,
                default_value=default_value))
    return target_dictionary


def now_milliseconds():
    return str(int(time.time() * 1000))


def two_weeks_ago():
    return str(int(time.time() * 1000 - (30 * (24 * 60 * 60 * 1000))))


def now_redshift():
    return datetime.datetime.now().strftime('%b %d,%Y  %H:%M:%S')


def now_datetime():
    return datetime.datetime.now()


def two_weeks_ago_datetime():
    return datetime.datetime.now() - datetime.timedelta(days=14)


def datetime_to_redshift(datetime_obj):
    return datetime_obj.strftime('%b %d,%Y  %H:%M:%S')


def datetime_to_milliseconds(datetime_obj):
    return int((datetime_obj - UNIX_EPOCH).total_seconds() * 1000)


class SafeMap(dict):
    def __missing__(self, key):
        return '{' + str(key) + '}'


class ListIndex:
    def __init__(self, index):
        self.index = index

    def __repr__(self):
        return 'ListIndex({index})'.format(index=str(self.index))

    def __eq__(self, other):
        return isinstance(other, (ListIndex,)) and self.index == other.index

    def __hash__(self):
        return self.index


def all_paths(thing, path=None, starting_path=None):
    path = path or tuple()
    starting_path = starting_path or []
    thing = get_value(thing, starting_path)

    if isinstance(thing, (dict,)):
        for key, value in thing.items():
            yield path + (key,)
            for i in all_paths(value, path=path + (key,)):
                yield i

    elif isinstance(thing, (list,)):
        for index, item in enumerate(thing):
            yield path + (ListIndex(index),)
            for i in all_paths(item, path=path + (ListIndex(index),)):
                yield i
    else:
        yield path


def last_element_list_index(path):
    return isinstance(path[-1], (ListIndex,))


def remove_list_indexes(path):
    return tuple(step for step in path if not isinstance(step, (ListIndex,)))


def matching_tail_paths(target_path, structure, starting_path=None):
    target_path = tuple(target_path)
    if starting_path is not None:
        starting_path = tuple(starting_path)
    seen = set()
    for path in all_paths(structure, starting_path=starting_path):
        if path in seen:
            continue
        else:
            seen.add(path)
        if last_element_list_index(path):
            continue

        temp_list = remove_list_indexes(path)

        if len(temp_list) < len(target_path):
            continue
        tail_of_path = temp_list[-1 * len(target_path):]
        if lists_equal(tail_of_path, target_path):
            logging.info('tail path: ' + str(target_path) + ' ' + str(path))
            yield (starting_path or tuple([])) + path


def lists_equal(l1, l2):
    return l1 == l2


def nevermind(thing):
    return thing


def replace_by_path(
    dictionary, target_path, target_value=None, function=None,
        function_args=None, function_kwargs=None, starting_path=None):

    function = function or nevermind
    function_args = function_args or tuple([])
    function_kwargs = function_kwargs or {}
    #dictionary_clone = copy.deepcopy(dictionary)
    dictionary_clone = dictionary
    for path in matching_tail_paths(
            target_path, dictionary_clone, starting_path=starting_path):
        current_value = get_value(dictionary, path)
        one_target_value = target_value or function(
            current_value, *function_args, **function_kwargs)
        set_value(dictionary, path, one_target_value)


def aggregate_values(dictionary, target_path, values=False):
    aggregated_values = []
    for path in matching_tail_paths(target_path, dictionary):
        current_value = get_value(dictionary, path)
        aggregated_values.append(
            list(current_value.values()) if values else current_value)
    logging.debug('aggregated_values: ' + str(aggregated_values))
    return aggregated_values if not values else aggregated_values[0]


class UberDict(dict):

    def rget(self, value):
        for i in meets_condition(self, lambda x: isinstance(x, (dict,))):
            for _key, _value in i.items():
                if _value == value:
                    yield _key


def iterate(thing, path=None, seen=None, start_path=None):
    '''
    TODO: allow specification of `start_path` to speed this up.
    '''
    hashed_thing = hash(str(thing))
    path = path or []
    start_path = start_path or []
    seen = seen or set([])
    if hashed_thing not in seen:
        if isinstance(thing, (list, tuple, dict,)):
            yield thing
        seen.add(hashed_thing)
        if isinstance(thing, (dict,)):
            for key, value in thing.items():
                for item in iterate(value, path=path + [key], seen=seen):
                    yield item
        elif isinstance(thing, (list, tuple,)):
            for index, item in enumerate(thing):
                for thingie in iterate(item, path=path + [ListIndex(index)], seen=seen):
                    yield thingie
        else:
            yield thing


def meets_condition(thing, func):
    for item in iterate(thing):
        if func(item):
            yield item




if __name__ == '__main__':
    import json
    d = json.load(open('./sample.json', 'r'))
    counter = 0
    while counter < 100:
        replace_by_path(d, ['vid'], target_value='hithere')
        counter += 1
        print(counter)
    import random
    logging.basicConfig(level=logging.INFO)
    replace_by_path(d, ['vid'], function=lambda x: str(x) + str(random.random()))



import setuptools

setuptools.setup(
    name="nanostream",
    version="0.1.8",
    author="Zachary Ernst",
    author_email="zac.ernst@gmail.com",
    description="Small-scale stream processing for ETL",
    url="https://github.com/zacernst/nanostream",
    scripts=['nanostream/bin/nanostream'],
    packages=[
        'nanostream',
        'nanostream.message',
        'nanostream.node_queue',
        'nanostream.node_classes',
        'nanostream.watchdog',
        'nanostream.utils',
        'nanostream.exp',],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
      'alabaster==0.7.11',
      'argh==0.26.2',
      'asn1crypto==0.24.0',
      'atomicwrites==1.2.1',
      'attrs==18.2.0',
      'Babel==2.6.0',
      'bcrypt==3.1.4',
      'behave==1.2.6',
      'boto3==1.7.62',
      'botocore==1.10.62',
      'certifi==2018.4.16',
      'cffi==1.11.5',
      'chardet==3.0.4',
      'civis==1.7.2',
      'click==6.7',
      'cloudpickle==0.5.3',
      'cryptography==2.3',
      'datadog==0.26.0',
      'decorator==4.3.0',
      'docutils==0.14',
      'graphviz==0.10.1',
      'gunicorn==19.9.0',
      'idna==2.7',
      'imagesize==1.1.0',
      'Jinja2==2.10',
      'jmespath==0.9.3',
      'joblib==0.11',
      'jsonref==0.1',
      'jsonschema==2.6.0',
      'kafka==1.3.1',
      'MarkupSafe==1.0',
      'prometheus_client',
      'more-itertools==4.3.0',
      'networkx==2.1',
      'packaging==18.0',
      'pandas==0.23.3',
      'paramiko==2.4.2',
      'parse==1.8.4',
      'parse-type==0.4.2',
      'pathlib2==2.3.2',
      'pathtools==0.1.2',
      'pkginfo==1.4.2',
      'pluggy==0.7.1',
      'pockets==0.6.2',
      'prettytable==0.7.2',
      'pubnub==4.1.0',
      'py==1.6.0',
      'pyasn1==0.4.3',
      'pycparser==2.18',
      'pycryptodomex==3.6.4',
      'Pygments==2.2.0',
      'PyNaCl==1.2.1',
      'pyparsing==2.2.1',
      'pytest==3.8.1',
      'python-dateutil==2.7.3',
      'python-dotenv==0.6.0',
      'pytz==2018.5',
      'pyyaml>=3.0',
      'requests==2.19.1',
      'requests-toolbelt==0.8.0',
      's3fs==0.1.5',
      's3transfer==0.1.13',
      'schedule==0.5.0',
      'six==1.11.0',
      'snowballstemmer==1.2.1',
      'Sphinx==1.8.1',
      'sphinx-rtd-theme==0.4.2',
      'sphinxcontrib-napoleon==0.6.1',
      'sphinxcontrib-websupport==1.1.0',
      'sqlparse==0.2.4',
      'TimedDict==0.2.4',
      'timeddictionary==0.1.1',
      'tqdm==4.23.4',
      'twine==1.11.0',
      'urllib3==1.23',
      'watchdog==0.8.3',
      'yapf==0.24.0']

)



"""
Node module
===========

The ``node`` module contains the ``NanoNode`` class, which is the foundation
for NanoStream.

"""

import time
import datetime
import uuid
import importlib
import logging
import os
import threading
import pprint
import sys
import copy
import random
import functools
import csv
import re
import io
import yaml
import types
import inspect
import prettytable
# import MySQLdb  #  Fix this for container script
import requests
import graphviz

from timed_dict.timed_dict import TimedDict
from nanostream.message.batch import BatchStart, BatchEnd
from nanostream.message.message import NanoStreamMessage
from nanostream.node_queue.queue import NanoStreamQueue
from nanostream.message.canary import Canary
from nanostream.message.poison_pill import PoisonPill
from nanostream.utils.set_attributes import set_kwarg_attributes
from nanostream.utils.data_structures import Row, MySQLTypeSystem
from nanostream.utils import data_structures as ds
from nanostream.utils.helpers import (
    replace_by_path, remap_dictionary, set_value, get_value, to_bool,
    aggregate_values)

from datadog import api, statsd, ThreadStats
from datadog import initialize as initialize_datadog

DEFAULT_MAX_QUEUE_SIZE = int(os.environ.get('DEFAULT_MAX_QUEUE_SIZE', 128))
MONITOR_INTERVAL = 1
STATS_COUNTER_MODULO = 4

PROMETHEUS = False

def no_op(*args, **kwargs):
    '''
    No-op function to serve as default ``get_runtime_attrs``.
    '''
    return None


def get_environment_variables(*args):
    '''
    Retrieves the environment variables listed in ``*args``.

    Args:
        args (list of str): List of environment variables.
    Returns:
        dict: Dictionary of environment variables to values. If the environment
            variable is not defined, the value is ``None``.
    '''
    environment_variables = args or []
    environment = {
        environment_variable: os.environ.get(environment_variable, None)
        for environment_variable in environment_variables}
    return environment


class bcolors:
    '''
    This class holds the values for the various colors that are used in the
    tables that monitor the status of the nodes.
    '''

    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class NothingToSeeHere:
    '''
    Vacuous class used as a no-op message type.
    '''
    pass


class Parameters:

    def __init__(self, **kwargs):
        self.parameters = kwargs


class NanoNode:
    '''
    The foundational class of `NanoStream`. This class is inherited by all
    nodes in a computation graph.

    Order of operations:
    1. Child class ``__init__`` function
    2. ``NanoNode`` ``__init__`` function
    3. ``preflight_function`` (Specified in initialization params)
    4. ``setup``
    5. start


    These methods have the following intended uses:

    1. ``__init__`` Sets attribute values and calls the ``NanoNode`` ``__init__``
       method.
    2. ``get_runtime_attrs`` Sets any attribute values that are to be determined
       at runtime, e.g. by checking environment variables or reading values
       from a database. The ``get_runtime_attrs`` should return a dictionary
       of attributes -> values, or else ``None``.
    3. ``setup`` Sets the state of the ``NanoNode`` and/or creates any attributes
       that require information available only at runtime.

    :ivar send_batch_markers: If ``True``, then a ``BatchStart`` marker will
        be sent when a new input is received, and a ``BatchEnd`` will be sent
        after the input has been processed. The intention is that a number of
        items will be emitted for each input received. For example, we might
        emit a table row-by-row for each input.
    :ivar get_runtime_attrs: A function that returns a dictionary-like object.
        The keys and values will be saved to this ``NanoNode`` object's
        attributes. The function is executed one time, upon starting the node.
    :ivar get_runtime_attrs_args: A tuple of arguments to be passed to the
        ``get_runtime_attrs`` function upon starting the node.
    :ivar get_runtime_attrs_kwargs: A dictionary of kwargs passed to the
        ``get_runtime_attrs`` function.
    :ivar runtime_attrs_destinations: If set, this is a dictionary mapping
        the keys returned from the ``get_runtime_attrs`` function to the
        names of the attributes to which the values will be saved.
    :ivar throttle: For each input received, a delay of ``throttle`` seconds
        will be added.
    :ivar keep_alive: If ``True``, keep the node's thread alive after
        everything has been processed.
    :ivar name: The name of the node. Defaults to a randomly generated hash.
        Note that this hash is not consistent from one run to the next.
    :ivar input_mapping: When the node receives a dictionary-like object,
        this dictionary will cause the keys of the dictionary to be remapped
        to new keys.
    :ivar retain_input: If ``True``, then combine the dictionary-like input
        with the output. If keys clash, the output value will be kept.
    :ivar input_message_keypath: Read the value in this keypath as the content
        of the incoming message.

    '''

    def __init__(
        self,
        *args,
        batch=False,
        get_runtime_attrs=no_op,
        get_runtime_attrs_args=None,
        get_runtime_attrs_kwargs=None,
        runtime_attrs_destinations=None,
        input_mapping=None,
        retain_input=True,
        throttle=0,
        keep_alive=True,
        max_errors=0,
        name=None,
        input_message_keypath=None,
        key=None,
        messages_received_counter=0,
        prefer_existing_value=False,
        messages_sent_counter=0,
        post_process_function=None,
        post_process_keypath=None,
        summary='',
        post_process_function_kwargs=None,
        output_key=None,
            **kwargs):
        self.name = name or uuid.uuid4().hex
        self.input_mapping = input_mapping or {}
        self.input_queue_list = []
        self.output_queue_list = []
        self.input_node_list = []
        self.input_message_keypath = input_message_keypath or []
        self.output_node_list = []
        self.global_dict = None  # We'll add a dictionary upon startup
        self.thread_dict = {}
        self.kill_thread = False
        self.prefer_existing_value = prefer_existing_value
        self.accumulator = {}
        self.output_key = output_key
        self.keep_alive = keep_alive
        self.retain_input = retain_input  # Keep the input dictionary and send it downstream
        self.throttle = throttle
        self.get_runtime_attrs = get_runtime_attrs
        self.get_runtime_attrs_args = get_runtime_attrs_args or tuple()
        self.get_runtime_attrs_kwargs = get_runtime_attrs_kwargs or {}
        self.runtime_attrs_destinations = runtime_attrs_destinations or {}
        self.key = key
        self.messages_received_counter = messages_received_counter
        self.messages_sent_counter = messages_sent_counter
        self.instantiated_at = datetime.datetime.now()
        self.started_at = None
        self.stopped_at = None
        self.finished = False
        self.error_counter = 0
        self.status = 'stopped'  # running, error, success
        self.max_errors = max_errors
        self.post_process_function_name = post_process_function  # Function to be run on result
        self.post_process_function_kwargs = post_process_function_kwargs or {}
        self.summary = summary
        self.prometheus_objects = None

        # Get post process function if one is named
        if self.post_process_function_name is not None:
            components = self.post_process_function_name.split('__')
            if len(components) == 1:
                module = None
                function_name = components[0]
                self.post_process_function = globals()[function_name]
            else:
                module = '.'.join(components[:-1])
                function_name = components[-1]
                module = importlib.import_module(module)
                self.post_process_function = getattr(module, function_name)
        else:
            self.post_process_function = None

        self.post_process_keypath = (
            post_process_keypath.split('.') if post_process_keypath is not None
            else None)

    def setup(self):
        '''
        To be overridden by child classes when we need to do something
        after setting attributes and the pre-flight function.
        '''
        logging.debug('No ``setup`` method for {class_name}.'.format(
            class_name=self.__class__.__name__))
        pass

    def __gt__(self, other):
        '''
        Convenience method so that we can link two nodes by ``node1 > node2``.
        '''
        self.add_edge(other)
        return other

    @property
    def is_source(self):
        '''
        Tests whether the node is a source or not, i.e. whether there are no
        inputs to the node.
        '''
        return len(self.input_queue_list) == 0

    @property
    def is_sink(self):
        '''
        Tests whether the node is a sink or not, i.e. whether there are no
        outputs from the node.
        '''
        return len(self.output_queue_list) == 0

    def add_edge(self, target, **kwargs):
        """
        Create an edge connecting `self` to `target`. The edge
        is really just a queue
        """
        max_queue_size = kwargs.get('max_queue_size', DEFAULT_MAX_QUEUE_SIZE)
        edge_queue = NanoStreamQueue(max_queue_size)

        self.output_node_list.append(target)
        target.input_node_list.append(self)

        edge_queue.source_node = self
        edge_queue.target_node = target

        # Make this recursive below
        # TODO: Add support for composition
        target.input_queue_list.append(edge_queue)
        self.output_queue_list.append(edge_queue)

    def start(self):
        '''
        Starts the node. This is called by ``NanoNode.global_start()``.
        '''

        # Run pre-flight function and save output if necessary
        self.started_at = datetime.datetime.now()
        logging.debug('Starting node: {node}'.format(
            node=self.__class__.__name__))
        # ``get_runtime_attrs`` returns a dict-like object whose keys and
        # values are stored as attributes of the ``NanoNode`` object.
        if self.get_runtime_attrs is not None:
            pre_flight_results = self.get_runtime_attrs(
                *self.get_runtime_attrs_args,
                **self.get_runtime_attrs_kwargs) or {}
            if self.runtime_attrs_destinations is not None:
                for key, value in pre_flight_results.items():
                    setattr(self, self.runtime_attrs_destinations[key], value)
            elif self.runtime_attrs_destinations is None:
                for key, value in pre_flight_results.items():
                    setattr(self, key, value)
            else:
                raise Exception(
                    'There is a ``get_runtime_attrs``, but the '
                    '``runtime_attrs_destinations`` is neither None nor a '
                    'dict-like object.')

        # We have to separate the pre-flight function, the setup of the
        # class, and any necessary startup functions (such as connecting
        # to a database).

        self.setup()  # Setup function?

        if self.is_source and not isinstance(self, (DynamicClassMediator, )):
            for output in self.generator():
                yield output, None
            self.finished = True
        else:
            logging.debug(
                'About to enter loop for reading input queue in {node}.'.format(
                    node=str(self)))
            while not self.finished:
                new_input_sentinal = False  # Set to ``True`` when we have a new input
                for input_queue in self.input_queue_list:
                    one_item = input_queue.get()
                    if one_item is None:
                        continue

                    # Keep track of where the message came from, useful for
                    # managing streaming joins, e.g.
                    message_source = input_queue.source_node

                    new_input_sentinal = True
                    self.messages_received_counter += 1

                    time.sleep(self.throttle)
                    logging.debug('Got item: {one_item}'.format(
                        one_item=str(one_item)))
                    # Get the content of a specific keypath, if one has
                    # been defined in the ``NanoNode`` initialization.
                    message_content = (
                        get_value(
                            one_item.message_content,
                            self.input_message_keypath)
                        if len(self.input_message_keypath) > 0
                        else one_item.message_content)
                    if (
                        isinstance(message_content, (dict,))
                        and len(message_content) == 1
                            and '__value__' in message_content):
                        message_content = message_content['__value__']
                    # If we receive a ``PoisonPill`` object, kill the thread.
                    if isinstance(message_content, (PoisonPill,)):
                        logging.debug('received poision pill.')
                        self.finished = True
                    # If we receive ``None``, then pass.
                    elif message_content is None:
                        pass
                    # Otherwise, process the message as usual, by calling
                    # the ``NanoNode`` object's ``process_item`` method.
                    else:
                        self.message = message_content
                        self.message_source = message_source
                        # Remap inputs; if input_mapping is a string,
                        # assume we're mapping to {input_mapping: value}, else
                        # assume we're renaming keys

                        if isinstance(self.input_mapping, (str,)):
                            self.message = {self.input_mapping: self.message}
                        elif (
                            isinstance(self.input_mapping, (dict,)) and
                            len(self.input_mapping) > 0 and
                                isinstance(self.message, (dict,))):
                            for key, value in list(self.message.items()):
                                if key in self.input_mapping:
                                    self.message[self.input_mapping[key]] = value
                        elif (
                            isinstance(self.input_mapping, (dict,)) and
                            len(self.input_mapping) > 0 and
                            hasattr(self.message, '__dict__') and
                                not isinstance(self.message, (dict,))):
                            for key, value in self.message.__dict__.items():
                                if key in self.input_mapping:
                                    self.message.__dict__[self.input_mapping[key]] = value
                        elif (
                            isinstance(self.input_mapping, (dict,)) and
                                len(self.input_mapping) > 0):
                            raise Exception('Bad case in input remapping.')
                        else:
                            pass

                        # Datadog
                        if self.datadog:
                            self.datadog_stats.increment(
                                '{pipeline_name}.{node_name}.input_counter'.format(
                                    pipeline_name=self.pipeline_name,
                                    node_name=self.name))

                        for output in self._process_item():
                            if self.datadog:
                                self.datadog_stats.increment(
                                    '{pipeline_name}.{node_name}.output_counter'.format(
                                        pipeline_name=self.pipeline_name,
                                        node_name=self.name))
                            yield output, one_item  # yield previous message
                if self.kill_thread:
                    break
                # Check input node(s) here to see if they're all ``.finished``
                this_node_finished = (
                    all(
                        node.finished for node in self.input_node_list) and
                    all(queue.empty for queue in self.input_queue_list))
                if this_node_finished:
                    self.finished = True
            self.cleanup()

    def cleanup(self):
        self.log_info('Cleanup called after shutdown.')

    def log_info(self, message=''):
        logging.info('{node_name}: {message}'.format(
            node_name=self.name, message=message))

    def terminate_pipeline(self):
        '''
        This method can be called on any node in a pipeline, and it will cause
        all of the nodes to terminate.
        '''
        for node in self.all_connected():
            if not node.finished:
                node.stopped_at = datetime.datetime.now()
                node.finished = True

    def process_item(self, *args, **kwargs):
        '''
        Default no-op for nodes.
        '''
        pass

    @property
    def __message__(self):
        '''
        If the node has an ``output_key`` defined, return the corresponding
        value in the message dictionary. If it does not, return the entire
        message dictionary.
        '''
        if self.key is None:
            out = self.message
        elif isinstance(self.key, (str,)):
            out = self.message[self.key]
        elif isinstance(self.key, (list,)):
            out = get_value(self.message, self.key)
        else:
            raise Exception('Bad type for input key.')
        return out

    def _process_item(self, *args, **kwargs):
        '''
        This extra indirection is so that we have a place to insert some
        error handling later.
        '''
        # Swap out the message if ``key`` is specified
        # If we're using prometheus, then increment a counter
        if self.prometheus_objects is not None:
            self.prometheus_objects['incoming_message_summary'].observe(random.random())
        message_arrival_time = time.time()

        try:
            for out in self.process_item(*args, **kwargs):
                if not isinstance(out, (dict, NothingToSeeHere)) and self.output_key is None:
                    logging.debug('Exception raised due to no key' + str(self.name))
                    raise Exception(
                        'Either message must be a dictionary or `output_key` '
                        'must be specified. {name}'.format(self.name))
                # Apply post_process_function if it's defined
                if self.post_process_function is not None:
                    set_value(
                        out,
                        self.post_process_keypath,
                        self.post_process_function(
                            get_value(
                                out,
                                self.post_process_keypath),
                            **self.post_process_function_kwargs))
                if self.prometheus_objects is not None:
                    self.prometheus_objects['outgoing_message_summary'].set(time.time() - message_arrival_time)

                yield out
        except Exception as err:
            self.error_counter += 1
            if self.error_counter > self.max_errors:
                logging.warning('message: ' + str(err.args) + str(self.__class__.__name__) + str(self.name))
                raise err
            else:
                logging.warning('oops')

    def processor(self):
        """
        This calls the user's ``process_item`` with just the message content,
        and then returns the full message.
        """
        logging.debug('Processing message content: {message_content}'.format(
            message_content=message.mesage_content))
        for result in self.process_item():
            result = NanoStreamMessage(result)
            yield result

    def stream(self):
        '''
        Called in each ``NanoNode`` thread.
        '''
        self.status = 'running'
        try:
            for output, previous_message in self.start():
                logging.debug('In NanoNode.stream.stream() --> ' + str(output))
                for output_queue in self.output_queue_list:
                    self.messages_sent_counter += 1
                    output_queue.put(
                        output,
                        block=True,
                        timeout=None,
                        previous_message=previous_message)
        except Exception as error:
            self.status = 'error'
            self.stopped_at = datetime.datetime.now()
            raise error
        self.status = 'success'
        self.stopped_at = datetime.datetime.now()

    @property
    def time_running(self):
        '''
        Return the number of wall-clock seconds elapsed since the node was
        started.
        '''

        if self.status == 'stopped':
            return None
        elif self.status == 'running':
            return datetime.datetime.now() - self.started_at
        elif self.stopped_at is None:
            return datetime.datetime.now() - self.started_at
        else:
            return self.stopped_at - self.started_at

    def all_connected(self, seen=None):
        '''
        Returns all the nodes connected (directly or indirectly) to ``self``.

        Args:
            seen (set): A set of all the nodes that have been identified as
                connected to ``self``.
        Returns:
            (set of ``NanoNode``): All the nodes connected to ``self``. This
                includes ``self``.
        '''
        seen = seen or set()

        if isinstance(self, (DynamicClassMediator, )):
            for node_name, node_dict in self.node_dict.items():
                node_obj = node_dict['obj']
                seen = seen | node_obj.all_connected(seen=seen)
        else:
            if self not in seen:
                seen.add(self)
            for node in self.input_node_list + self.output_node_list:
                if node in seen:
                    continue
                seen.add(node)
                seen = seen | node.all_connected(seen=seen)
        return seen

    def broadcast(self, broadcast_message):
        '''
        Puts the message into all the input queues for all connected nodes.
        '''
        for node in self.all_connected():
            for input_queue in node.input_queue_list:
                input_queue.put(broadcast_message)

    def global_start(
        self, datadog=False, prometheus=False, pipeline_name=None):
        '''
        Starts every node connected to ``self``.
        '''

        def prometheus_init():
            from prometheus_client import start_http_server, Summary, Gauge, Histogram, Counter
            for node in self.all_connected():
                node.prometheus_objects = {}
                summary = Summary(node.name + '_incoming', 'Summary of incoming messages')
                node.prometheus_objects['incoming_message_summary'] = summary
                node.prometheus_objects['outgoing_message_summary'] = Gauge(node.name + '_outgoing', 'Summary of outgoing messages')
            start_http_server(8000)

        if PROMETHEUS:
            prometheus_init()

        # thread_dict = self.thread_dict
        global_dict = {}

        # Initialize datadog if we're doing that
        datadog_stats = ThreadStats() if datadog else None
        if datadog:
            datadog_options = {
                'api_key': os.environ['DATADOG_API_KEY'],
                'app_key': os.environ['DATADOG_APP_KEY']}
            initialize_datadog(**datadog_options)
            datadog_stats.start()

        run_id = uuid.uuid4().hex
        for node in self.all_connected():
            # Set the pipeline name on the attribute of each node
            node.pipeline_name = pipeline_name or uuid.uuid4().hex
            node.datadog_stats = datadog_stats
            # Set a unique run_id
            node.run_id = run_id
            # Tell each node whether they're logging to datadog
            node.datadog = datadog
            node.global_dict = global_dict  # Establishing shared globals
            logging.info('global_start:' + str(self))
            thread = threading.Thread(target=NanoNode.stream, args=(node, ))
            thread.start()
            node.thread_dict = self.thread_dict
            self.thread_dict[node.name] = thread
        monitor_thread = threading.Thread(
            target=NanoNode.thread_monitor, args=(self,), daemon=True)
        monitor_thread.start()

    @property
    def input_queue_size(self):
        '''
        Return the total number of items in all of the queues that are inputs
        to this node.
        '''

        return sum(
            [input_queue.queue.qsize() for input_queue in self.input_queue_list])

    def kill_pipeline(self):
        for node in self.all_connected():
            node.finished = True

    def draw_pipeline(self):
        '''
        Draw the pipeline structure using graphviz.
        '''

        dot = graphviz.Digraph()
        for node in self.all_connected():
            dot.node(node.name, node.name, shape='box')
        for node in self.all_connected():
            for target_node in node.output_node_list:
                dot.edge(node.name, target_node.name)
        dot.render('pipeline_drawing.gv', view=True)

    def thread_monitor(self):
        '''
        This function loops over all of the threads in the pipeline, checking
        that they are either ``finished`` or ``running``. If any have had an
        abnormal exit, terminate the entire pipeline.
        '''

        counter = 0
        pipeline_finished = all(
            node.finished for node in self.all_connected())
        error = False
        while not pipeline_finished:
            time.sleep(MONITOR_INTERVAL)
            counter += 1

            # Check whether all the workers have ``.finished``
            pipeline_finished = all(
                node.finished for node in self.all_connected())

            if counter % STATS_COUNTER_MODULO == 0:
                table = prettytable.PrettyTable(
                    ['Node', 'Class', 'Received', 'Sent',
                     'Queued', 'Status', 'Time'])
                for node in self.all_connected():
                    if node.status == 'running':
                        status_color = bcolors.WARNING
                    elif node.status == 'stopped':
                        status_color = ''
                    elif node.status == 'error':
                        status_color = bcolors.FAIL
                        error = True
                    elif node.status == 'success':
                        status_color = bcolors.OKGREEN
                    else:
                        assert False

                    table.add_row(
                        [
                            node.name,
                            node.__class__.__name__,
                            node.messages_received_counter,
                            node.messages_sent_counter,
                            node.input_queue_size,
                            status_color + node.status + bcolors.ENDC,
                            node.time_running])
                logging.info('\n' + str(table))
                if error:
                    logging.info('Terminating due to error.')
                    self.terminate_pipeline()
                    pipeline_finished = True
                    break

            pipeline_finished = all(
                node.finished for node in self.all_connected())

        logging.info('Pipeline finished.')
        logging.info('Sending terminate signal to nodes.')
        logging.info('Messages that are being processed will complete.')

        sys.exit(0)


class CounterOfThings(NanoNode):
    def foo__init__(self, *args, start=0, end=None, **kwargs):
        self.start = start
        self.end = end
        super(CounterOfThings, self).__init__(*args, **kwargs)

    def generator(self):
        '''
        Just start counting integers
        '''
        counter = 1
        while 1:
            yield counter
            counter += 1
            if counter > 10:
                assert False


class InsertData(NanoNode):

    def __init__(self, overwrite=True, overwrite_if_null=True, value_dict=None, **kwargs):
        self.overwrite = overwrite
        self.overwrite_if_null = overwrite_if_null
        self.value_dict = value_dict or {}
        super(InsertData, self).__init__(**kwargs)

    def process_item(self):
        for key, value in self.value_dict.items():
            if (key not in self.__message__) or self.overwrite or (
                self.__message__.get(key) == None and
                     self.overwrite_if_null):
                self.__message__[key] = value
        yield self.__message__


class RandomSample(NanoNode):
    '''
    Lets through only a random sample of incoming messages. Might be useful
    for testing, or when only approximate results are necessary.
    '''

    def __init__(self, sample=.1):
        self.sample = sample

    def process_item(self):
        yield self.message if random.random() <= self.sample else None


class SubstituteRegex(NanoNode):
    def __init__(self, match_regex=None, substitute_string=None, *args, **kwargs):
        self.match_regex = match_regex
        self.substitute_string = substitute_string
        self.regex_obj = re.compile(self.match_regex)
        super(SubstituteRegex, self).__init__(*args, **kwargs)

    def process_item(self):
        out = self.regex_obj.sub(self.substitute_string, self.message[self.key])
        yield out


class CSVToDictionaryList(NanoNode):
    def __init__(self, **kwargs):
        super(CSVToDictionaryList, self).__init__(**kwargs)

    def process_item(self):
        csv_file_obj = io.StringIO(self.__message__)
        csv_reader = csv.DictReader(csv_file_obj)
        output = [row for row in csv_reader]
        yield output


class SequenceEmitter(NanoNode):
    '''
    Emits ``sequence`` ``max_sequences`` times, or forever if
    ``max_sequences`` is ``None``.
    '''

    def __init__(self, sequence, *args, max_sequences=1, **kwargs):
        self.sequence = sequence
        self.max_sequences = max_sequences
        super(SequenceEmitter, self).__init__(*args, **kwargs)

    def generator(self):
        '''
        Emit the sequence ``max_sequences`` times.
        '''
        type_dict = {
            'int': int,
            'integer': int,
            'str': str,
            'string': str,
            'float': float,
            'bool': to_bool}
        counter = 0
        while counter < self.max_sequences:
            for item in self.sequence:
                if isinstance(item, (dict,)) and 'value' in item and 'type' in item:
                    item = type_dict[item['type'].lower()](item['value'])
                item = {self.output_key: item}
                yield item
            counter += 1

    def process_item(self):
        '''
        Emit the sequence ``max_sequences`` times.
        '''
        type_dict = {
            'int': int,
            'integer': int,
            'str': str,
            'string': str,
            'float': float,
            'bool': to_bool}
        counter = 0
        while counter < self.max_sequences:
            for item in self.sequence:
                if isinstance(item, (dict,)) and 'value' in item and 'type' in item:
                    item = type_dict[item['type'].lower()](item['value'])
                item = {self.output_key: item}
                yield item
            counter += 1


class GetEnvironmentVariables(NanoNode):
    def __init__(self, mappings=None, environment_variables=None, **kwargs):
        self.environment_mappings = mappings or {}
        self.environment_variables = environment_variables or []
        super(GetEnvironmentVariables, self).__init__(**kwargs)

    def generator(self):
        environment = {
            self.environment_mappings.get(
                environment_variable, environment_variable): os.environ.get(
                    environment_variable, None)
                for environment_variable in self.environment_variables}
        yield environment

    def process_item(self):
        environment = {
            self.environment_mappings.get(
                environment_variable, environment_variable): os.environ.get(
                    environment_variable, None)
                for environment_variable in self.environment_variables}
        yield environment


class SimpleTransforms(NanoNode):
    def __init__(self, missing_keypath_action='ignore', starting_path=None,
            transform_mapping=None, target_value=None, keypath=None, **kwargs):

        self.missing_keypath_action = missing_keypath_action
        self.transform_mapping = transform_mapping or []
        self.functions_dict = {}
        self.starting_path = starting_path

        for transform in self.transform_mapping:
            # Not doing the transforms; only loading the right functions here
            function_name = transform.get('target_function', None)
            full_function_name = function_name
            if function_name is not None:
                components = function_name.split('__')
                if len(components) == 1:
                    module = None
                    function_name = components[0]
                    function_obj = globals()[function_name]
                else:
                    module = '.'.join(components[:-1])
                    function_name = components[-1]
                    module = importlib.import_module(module)
                    function = getattr(module, function_name)
                self.functions_dict[full_function_name] = function

        super(SimpleTransforms, self).__init__(**kwargs)

    def process_item(self):
        logging.info('TRANSFORM ' + str(self.name))
        logging.info(self.name + ' ' + str(self.message))
        for transform in self.transform_mapping:
            path = transform['path']
            target_value = transform.get('target_value', None)
            function_name = transform.get('target_function', None)
            starting_path = transform.get('starting_path', None)
            if function_name is not None:
                function = self.functions_dict[function_name]
            else:
                function = None
            function_kwargs = transform.get('function_kwargs', None)
            function_args = transform.get('function_args', None)
            logging.info(self.name + ' calling replace_by_path:')
            replace_by_path(
                self.message,
                tuple(path),
                target_value=target_value,
                function=function,
                function_args=function_args,
                starting_path=starting_path,
                function_kwargs=function_kwargs)
            logging.info('after SimpleTransform: ' + self.name + str(self.message))
        yield self.message


class Serializer(NanoNode):
    '''
    Takes an iterable thing as input, and successively yields its items.
    '''

    def __init__(self, values=False, *args, **kwargs):
        self.values = values
        super(Serializer, self).__init__(**kwargs)

    def process_item(self):
        if self.values:
            for item in self.__message__.values():
                yield item
        else:
            for item in self.__message__:
                logging.info(self.name + ' ' + str(item))
                yield item


class AggregateValues(NanoNode):
    '''
    Does that.
    '''

    def __init__(self, values=False, tail_path=None, **kwargs):
        self.tail_path = tail_path
        self.values = values
        super(AggregateValues, self).__init__(**kwargs)

    def process_item(self):
        values = aggregate_values(self.__message__, self.tail_path, values=self.values)
        logging.info('aggregate_values ' + self.name + ' ' + str(values))
        yield values


class Filter(NanoNode):
    '''
    Applies tests to each message and filters out messages that don't pass

    Built-in tests:
        key_exists
        value_is_true
        value_is_not_none

    Example:
        {'test': 'key_exists',
         'key': mykey}

    '''
    def __init__(self, test=None, test_keypath=None, value=True, *args, **kwargs):
        self.test = test
        self.value = value
        self.test_keypath = test_keypath or []
        super(Filter, self).__init__(*args, **kwargs)

    @staticmethod
    def _key_exists(message, key):
        return key in message

    @staticmethod
    def _value_is_not_none(message, key):
        return get_value(message, key) is not None

    @staticmethod
    def _value_is_true(message, key):
        return to_bool(message.get(key, False))

    def process_item(self):
        if self.test in ['key_exists', 'value_is_not_none', 'value_is_true']:
            result = (
                getattr(self, '_' + self.test)(
                    self.__message__, self.test_keypath) == self.value)
        else:
            raise Exception('Unknown test: {test_name}'.format(
                test_name=test))
        if result:
            logging.info('Sending message through')
            yield self.message
        else:
            logging.info('Blocking message: ' + str(self.__message__))
            yield NothingToSeeHere()


class StreamMySQLTable(NanoNode):
    def __init__(
        self,
        *args,
        host='localhost',
        user=None,
        table=None,
        password=None,
        database=None,
        port=3306,
        to_row_obj=False,
        send_batch_markers=True,
            **kwargs):
        self.host = host
        self.user = user
        self.to_row_obj = to_row_obj
        self.password = password
        self.database = database
        self.port = port
        self.table = table

        super(StreamMySQLTable, self).__init__(**kwargs)

    def setup(self):
        self.db = MySQLdb.connect(
            passwd=self.password,
            db=self.database,
            user=self.user,
            port=self.port)
        self.cursor = MySQLdb.cursors.DictCursor(self.db)

        self.table_schema_query = ('''SELECT column_name, column_type '''
                                   '''FROM information_schema.columns '''
                                   '''WHERE table_name='{table}';'''.format(
                                       table=self.table))

        self.table_schema = self.get_schema()
        # Need a mapping from header to MYSQL TYPE
        for mapping in self.table_schema:
            column = mapping['column_name']
            type_string = mapping['column_type']
            this_type = ds.MySQLTypeSystem.type_mapping(type_string)
            # Start here:
            #    store the type_mapping
            #    use it to cast the data into the MySQLTypeSchema
            #    ensure that the generator is emitting MySQLTypeSchema objects

    def get_schema(self):
        self.cursor.execute(self.table_schema_query)
        table_schema = self.cursor.fetchall()
        return table_schema

    def generator(self):
        if self.send_batch_markers:
            yield BatchStart(schema=self.table_schema)
        self.cursor.execute(
            """SELECT * FROM {table};""".format(table=self.table))
        result = self.cursor.fetchone()
        while result is not None:
            if self.to_row_obj:
                result = Row.from_dict(result, type_system=MySQLTypeSystem)
            yield result
            result = self.cursor.fetchone()
        if self.send_batch_markers:
            yield BatchEnd()


class PrinterOfThings(NanoNode):

    @set_kwarg_attributes()
    def __init__(self, disable=False, pretty=False,
            prepend='printer: ', **kwargs):
        self.disable = disable
        self.pretty = pretty
        super(PrinterOfThings, self).__init__(**kwargs)
        logging.debug('Initialized printer...')

    def process_item(self):
        print(self.prepend)
        if self.pretty:
            pprint.pprint(self.__message__, indent=2)
        else:
            print(str(self.__message__))
        print('\n')
        print('------------')
        if random.random() < -.1:
            print('!!!!!!!!!!!')
            assert False
        yield self.message


class ConstantEmitter(NanoNode):
    '''
    Send a thing every n seconds
    '''

    def __init__(self, thing=None, delay=2, **kwargs):
        self.thing = thing
        self.delay = delay
        super(ConstantEmitter, self).__init__(**kwargs)

    def generator(self):
        while 1:
            if random.random() < -.1:
                assert False
            time.sleep(self.delay)
            yield self.thing


class TimeWindowAccumulator(NanoNode):
    '''
    Every N seconds, put the latest M seconds data on the queue.
    '''

    @set_kwarg_attributes()
    def __init__(self, time_window=None, send_interval=None, **kwargs):
        pass


class LocalFileReader(NanoNode):
    @set_kwarg_attributes()
    def __init__(self,
                 directory='.',
                 send_batch_markers=True,
                 serialize=False,
                 read_mode='r',
                 **kwargs):
        super(LocalFileReader, self).__init__(**kwargs)

    def process_item(self):
        filename = '/'.join([self.directory, self.message])
        with open(filename, self.read_mode) as file_obj:
            if self.serialize:
                if self.send_batch_markers:
                    yield BatchStart()
                for line in file_obj:
                    output = line
                    yield output
                yield BatchEnd()
            else:
                output = file_obj.read()
                yield output


class CSVReader(NanoNode):
    @set_kwarg_attributes()
    def __init__(self, send_batch_markers=True, to_row_obj=True, **kwargs):
        super(CSVReader, self).__init__(**kwargs)

    def process_item(self):
        file_obj = io.StringIO(self.message)
        reader = csv.DictReader(file_obj)
        if self.send_batch_markers:
            yield BatchStart()
        for row in reader:
            if self.to_row_obj:
                row = Row.from_dict(row)
            yield row
        if self.send_batch_markers:
            yield BatchEnd()


class LocalDirectoryWatchdog(NanoNode):
    def __init__(self, directory='.', check_interval=3, **kwargs):
        self.directory = directory
        self.latest_arrival = time.time()
        self.check_interval = check_interval
        super(LocalDirectoryWatchdog, self).__init__(**kwargs)

    def generator(self):
        while self.keep_alive:
            logging.debug('sleeping...')
            time.sleep(self.check_interval)
            time_in_interval = None
            for filename in os.listdir(self.directory):
                last_modified_time = os.path.getmtime('/'.join(
                    [self.directory, filename]))
                if last_modified_time > self.latest_arrival:
                    yield '/'.join([self.directory, filename])
                    if (time_in_interval is None
                            or last_modified_time > time_in_interval):
                        time_in_interval = last_modified_time
                        logging.debug('time_in_interval: ' +
                                      str(time_in_interval))
            if time_in_interval is not None:
                self.latest_arrival = time_in_interval


class StreamingJoin(NanoNode):
    '''
    Joins two streams on a key, using exact match only. MVP.
    '''

    def __init__(self, window=30, streams=None, *args, **kwargs):
        self.window = window
        self.streams = streams
        self.stream_paths = streams
        self.buffers = {
            stream_name: TimedDict(timeout=self.window)
            for stream_name in self.stream_paths.keys()}
        super(StreamingJoin, self).__init__(*args, **kwargs)

    def process_item(self):
        '''
        '''
        value_to_match = get_value(
            self.message, self.stream_paths[self.message_source.name])
        # Check for matches in all other streams.
        # If complete set of matches, yield the merged result
        # If not, add it to the `TimedDict`.
        yield('hi')


class DynamicClassMediator(NanoNode):
    def __init__(self, *args, **kwargs):

        super(DynamicClassMediator, self).__init__(**kwargs)

        for node_name, node_dict in self.node_dict.items():
            cls_obj = node_dict['cls_obj']
            node_obj = cls_obj(**kwargs)
            node_dict['obj'] = node_obj

        for edge in self.raw_config['edges']:
            source_node_obj = self.node_dict[edge['from']]['obj']
            target_node_obj = self.node_dict[edge['to']]['obj']
            source_node_obj > target_node_obj

        def bind_methods():
            for attr_name in dir(DynamicClassMediator):
                if attr_name.startswith('_'):
                    continue
                attr_obj = getattr(DynamicClassMediator, attr_name)
                if not isinstance(attr_obj, types.FunctionType):
                    continue
                setattr(self, attr_name, types.MethodType(attr_obj, self))

        bind_methods()

        source = self.get_source()
        self.input_queue_list = source.input_queue_list
        sink = self.get_sink()
        self.output_queue_list = sink.output_queue_list

        self.output_node_list = sink.output_node_list
        self.input_node_list = source.input_node_list

    def get_sink(self):
        sinks = self.sink_list()
        if len(sinks) > 1:
            raise Exception(
                '`DynamicClassMediator` may have no more than one sink.')
        elif len(sinks) == 0:
            return None
        return sinks[0]

    def get_source(self):
        sources = self.source_list()
        if len(sources) > 1:
            raise Exception(
                '`DynamicClassMediator` may have no more than one source.')
        elif len(sources) == 0:
            return None
        return sources[0]

    def sink_list(self):
        sink_nodes = []
        for node_name, node_dict in self.node_dict.items():
            node_obj = node_dict['obj']
            if len(node_obj.output_queue_list) == 0:
                sink_nodes.append(node_obj)
        return sink_nodes

    def source_list(self):
        source_nodes = [
            node_dict['obj'] for node_dict in self.node_dict.values()
            if node_dict['obj'].is_source]
        return source_nodes

    def hi(self):
        return 'hi'


def get_node_dict(node_config):
    node_dict = {}
    for node_config in node_config['nodes']:
        node_class = globals()[node_config['class']]
        node_name = node_config['name']
        node_dict[node_name] = {}
        node_dict[node_name]['class'] = node_class
        frozen_arguments = node_config.get('frozen_arguments', {})
        node_dict[node_name]['frozen_arguments'] = frozen_arguments
        node_obj = node_class(**frozen_arguments)
        # node_dict[node_name]['obj'] = node_obj
        node_dict[node_name]['remapping'] = node_config.get('arg_mapping', {})
    return node_dict


def kwarg_remapper(f, **kwarg_mapping):
    reverse_mapping = {value: key for key, value in kwarg_mapping.items()}
    logging.debug('kwarg_mapping:' + str(kwarg_mapping))
    parameters = [i for i, _ in list(inspect.signature(f).parameters.items())]
    for kwarg in parameters:
        if kwarg not in kwarg_mapping:
            reverse_mapping[kwarg] = kwarg

    def remapped_function(*args, **kwargs):
        remapped_kwargs = {}
        for key, value in kwargs.items():
            if key in reverse_mapping:
                remapped_kwargs[reverse_mapping[key]] = value
        logging.debug('renamed function with kwargs: ' + str(remapped_kwargs))

        return f(*args, **remapped_kwargs)

    return remapped_function


def template_class(class_name, parent_class, kwargs_remapping,
                   frozen_arguments_mapping):

    kwargs_remapping = kwargs_remapping or {}
    frozen_init = functools.partial(parent_class.__init__,
                                    **frozen_arguments_mapping)
    if isinstance(parent_class, (str, )):
        parent_class = globals()[parent_class]
    cls = type(class_name, (parent_class, ), {})
    setattr(cls, '__init__', kwarg_remapper(frozen_init, **kwargs_remapping))
    return cls


def class_factory(raw_config):
    new_class = type(raw_config['name'], (DynamicClassMediator, ), {})
    new_class.node_dict = get_node_dict(raw_config)
    new_class.class_name = raw_config['name']
    new_class.edge_list_dict = raw_config.get('edges', [])
    new_class.raw_config = raw_config

    for node_name, node_config in new_class.node_dict.items():
        _class = node_config['class']
        cls = template_class(node_name, _class, node_config['remapping'],
                             node_config['frozen_arguments'])
        setattr(cls, 'raw_config', raw_config)
        node_config['cls_obj'] = cls
    # Inject?
    globals()[new_class.__name__] = new_class
    return new_class


class Remapper(NanoNode):

    def __init__(self, mapping=None, **kwargs):
        self.remapping_dict = mapping or {}
        super(Remapper, self).__init__(**kwargs)

    def process_item(self):
        logging.info('Remapper {node}:'.format(node=self.name) + str(self.__message__))
        out = remap_dictionary(self.__message__, self.remapping_dict)
        yield out


class BatchMessages(NanoNode):
    def __init__(
        self, batch_size=None, batch_list=None,
            counter=0, timeout=5, **kwargs):
        self.batch_size = batch_size
        self.timeout = timeout
        self.counter = 0
        self.batch_list = batch_list or []
        super(BatchMessages, self).__init__(**kwargs)

    def process_item(self):
        self.counter += 1
        self.batch_list.append(self.__message__)
        logging.info(self.name + ' ' + str(self.__message__))
        out = NothingToSeeHere()
        if self.counter % self.batch_size == 0:
            out = self.batch_list
            logging.debug('BatchMessages: ' + str(out))
            self.batch_list = []
        yield out

    def cleanup(self):
        yield self.batch_list


if __name__ == '__main__':
    pass



"""
Classes that can be instantiated in a ``NanoStreamGraph`` to
read and write files.
"""

import random
import hashlib
from nanostream_processor import *


class FileReader(NanoStreamProcessor):
    """
    Reads a file and passes its contents down the pipeline.
    """

    def process_item(self):
        """
        Reads a file into memory.

        Args:
            item (str): Filename
        """

        with open(self.message, 'r') as nano_file:
            contents = nano_file.read()
        return contents


def random_hash(seed=None):
    """
    Returns a random hash; usually for naming files uniquely.
    """

    return hashlib.md5(seed or str(random.random())).hexdigest()


class FileWriter(NanoStreamProcessor):
    """
    Writes the message to a file with a random name in a specified
    directory.
    """

    def __init__(self, path=None):
        """
        Initializer.

        Args:
            path (str): Directory path for the file.
        """

        self.path = path or '.'
        super(FileWriter, self).__init__()

    def process_item(self):
        """
        Writes the file to the ``self.path`` directory, with a random
        filename generated by the ``random_hash`` function.
        """
        with open('/'.join([self.path, random_hash()]), 'w') as nano_file:
            print('file writer: ', self.message)
            nano_file.write(self.message)
        return self.message



import logging
import types
import hashlib
from nanostream.utils.helpers import *


logging.basicConfig(level=logging.ERROR)


def cast_generators(some_dict):
    if isinstance(some_dict, (dict,)):
        for key, value in some_dict.items():
            if isinstance(value, (types.GeneratorType,)):
                some_dict[key] = list(value)


class TreeHorn:
    pass


class Label:

    def __init__(self, label):
        self.label = label
        self.treehorns = []

    def __le__(self, other):
        if isinstance(other, (TreeHorn,)):
            self.treehorns.append(other)
        else:
            raise Exception('Right side of Label assignment must be TreeHorn.')

    def __repr__(self):
        return 'Label({label})'.format(label=self.label)

    def apply(self, generator):
        for node in generator():
            generator.labels.add(self)

    def __eq__(self, other):
        return self.label == other.label

    def __hash__(self):
        return int(hashlib.md5(bytes(self.label, 'utf8')).hexdigest(), 16)


class GoSomewhere(TreeHorn):

    def __init__(self, condition=None, **kwargs):
        self.condition = condition
        self.label = None
        super(GoSomewhere, self).__init__(**kwargs)

    def __call__(self, thing):
        generator = (
            thing.descendants if self.direction == 'down' else thing.ancestors)
        for node in generator():
            logging.info('go somewhere: ' + str(node))
            if self.condition(node) == self.condition.truth_value:
                yield node
                if self.label is not None:
                    node.labels.add(self.label)

    def apply_label(self, label):
        self.label = label
        return self


class GoDown(GoSomewhere):
    def __init__(self, **kwargs):
        self.direction = 'down'
        super(GoDown, self).__init__(**kwargs)


class GoUp(GoSomewhere):
    def __init__(self, **kwargs):
        self.direction = 'up'
        super(GoDown, self).__init__(**kwargs)


class MeetsCondition(TreeHorn):

    def __init__(
        self, test_function=None,
            truth_value=True, **kwargs):
        '''
        Fix generator in __call__ so that it depends on whether the argument
        is a generator or a TracedObject.
        '''
        self.test_function = test_function or (lambda x: True)
        self.truth_value = truth_value

    def _test(self, node):
        return self.test_function(node.value)

    @staticmethod
    def _trivial_generator(thing):
        yield thing

    def __call__(self, thing):
        return self.test_function(thing)

    def __and__(self, other):
        return And(self, other)

    def __invert__(self):
        return Not(self)

    def __or__(self, other):
        return ~ (~ self & ~ other)

    def __eq__(self, other):
        return (self & other) | (~ self & ~ other)

    def __ne__(self, other):
        return ~ (self == other)




class HasDescendantOrAncestor(MeetsCondition):

    def __init__(self, condition=None, **kwargs):
        self.condition = condition
        super(HasDescendantOrAncestor, self).__init__(**kwargs)

    def __call__(self, thing):
        generator = (
            thing.descendants() if isinstance(self, (HasDescendant,)) else
            thing.ancestors())
        for outer_node in generator:
            sentinal = False
            for _ in self.condition(outer_node):
                sentinal = True
                break
            if sentinal:
                yield outer_node


class HasDescendant(HasDescendantOrAncestor):

    def __init__(self, condition=None, **kwargs):
        self.condition = condition
        super(HasDescendant, self).__init__(
            condition=condition, **kwargs)

    def __call__(self, thing):
        return any(self.condition(node) for node in thing.descendants())


class HasAncestor(HasDescendantOrAncestor):

    def __init__(self, condition=None, **kwargs):
        self.condition = condition
        super(HasAncestor, self).__init__(
            condition=condition, **kwagrs)

    def __call__(self, thing):
        return any(self.condition(node) for node in thing.descendants())


class HasKey(MeetsCondition):
    def __init__(self, key=None, **kwargs):
        self.key = key

        def _condition(thing):
            return isinstance(thing, (TracedDictionary,)) and self.key in thing

        super(HasKey, self).__init__(test_function=_condition, **kwargs)


class And(MeetsCondition):

    def __init__(self, conjunct_1, conjunct_2, **kwargs):
        self.conjunct_1 = conjunct_1
        self.conjunct_2 = conjunct_2

        def _test_function(node):
            return (
                self.conjunct_1.test_function(node) and
                self.conjunct_2.test_function(node))

        super(And, self).__init__(test_function=_test_function, **kwargs)


class Not(MeetsCondition):

    def __init__(self, condition, **kwargs):
        self.condition = condition

        def _test_function(node):
            return not self.condition.test_function(node)

        super(Not, self).__init__(test_function=_test_function, **kwargs)


class IsList(MeetsCondition):

    def __init__(self, direction='down', **kwargs):
        super(IsList, self).__init__(
            test_function=(lambda x: isinstance(x, (TracedList,))), **kwargs)


class IsDictionary(MeetsCondition):

    def __init__(self, direction='down', **kwargs):
        super(IsDictionary, self).__init__(
            direction=direction,
            test_function=(lambda x: isinstance(x, (TracedDictionary,))))


class TracedObject:

    def __init__(
        self,
        path=None,
        parent=None,
        parent_key=None,
            parent_list_index=None):
        self.path = path or []
        self.parent = parent
        self.parent_key = parent_key
        self.labels = set()
        self.children = []
        self.parent_list_index = parent_list_index

    def enumerate(self):
        return self.root.descendants()

    def __eq__(self, other):
        return self.to_python() == other.to_python()

    def to_python(self):
        '''
        Disfrobulates the object back into Python primitive types.
        '''
        if isinstance(self, (TracedDictionary,)):
            out = {key: value.to_python() for key, value in self.items()}
        elif isinstance(self, (TracedList,)):
            out = [i.to_python() for i in self]
        elif isinstance(self, (TracedPrimitive,)):
            out = self.thing
        return out

    @property
    def is_root(self):
        return self.parent is None

    @property
    def is_leaf(self):
        return

    def ancestors(self, include_self=False):
        if include_self:
            yield self
        if self.parent is not None:
            yield self.parent
            for i in  self.parent.ancestors(include_self=False):
                yield i

    def descendants(self, include_self=False):
        if include_self:
            yield self
        for child in self.children:
            yield child
            for grandchild in child.descendants(include_self=False):
                yield grandchild

    @property
    def root(self):
        if self.is_root:
            return self
        else:
            return self.parent.root

    def __repr__(self):
        return self.thing.__repr__()


class ListIndex:

    def __init__(self, index):
        self.index = index

    def __repr__(self):
        return 'ListIndex({index})'.format(index=self.index)

    def __eq__(self, other):
        return self.index == other.index

class TracedList(TracedObject, list):

    def __init__(
        self,
        *args,
        path=None,
        parent=None,
        parent_key=None,
            parent_list_index=None):

        super(TracedList, self).__init__(
            path=path,
            parent=parent,
            parent_key=parent_key,
            parent_list_index=parent_list_index)
        for index, item in enumerate(args):
            child = splitter(
                item,
                path=self.path + [ListIndex(index)],
                parent=self,
                parent_key=None,
                parent_list_index=ListIndex(index))
            self.children.append(child)
            self.append(child)

    def __repr__(self):
        return '[' + ', '.join([item.__repr__() for item in self]) + ']'


class PathEndsIn(MeetsCondition):
    def __init__(self, path=None, **kwargs):
        self.path = path or []
        _test_function = (lambda x: x.path[-1 * len(self.path):] == self.path)
        super(PathEndsIn, self).__init__(test_function=_test_function, **kwargs)


def splitter(
    thing,
    path=None,
    parent=None,
    parent_key=None,
        parent_list_index=None):

    if isinstance(thing, (dict,)):
        return TracedDictionary(
            thing, path=path, parent=parent, parent_key=parent_key)
    elif isinstance(thing, (list, tuple,)):
        return TracedList(
            *thing, path=path, parent=parent, parent_key=parent_key)
    else:
        return TracedPrimitive(
            thing, path=path, parent=parent, parent_key=parent_key)


class TracedPrimitive(TracedObject):

    def __init__(self, thing, path=None, parent=None, parent_key=None):

        self.thing = thing
        super(TracedPrimitive, self).__init__(
            path=path, parent=parent, parent_key=parent_key)


class TracedDictionary(TracedObject, dict):

    def __init__(
        self,
        thing,
        path=None,
        parent=None,
        parent_key=None,
            parent_list_index=None):

        self.thing = thing
        super(
            TracedDictionary, self).__init__(
                path=path, parent=parent, parent_key=parent_key)
        for key, value in thing.items():
            child = splitter(
                value, parent=self, parent_key=key, path=self.path + [key])
            self[key] = child
            self.children.append(child)


if __name__ == '__main__':
    d = {
        'foo': 'bar',
        'bar': 'baz',
        'baz': {'foobar': 1, 'goober': 2},
        'qux': ['foo', 'foobarbaz', 'ding', {'goo': 'blergh'}],
        'a': {'b': {'c': 'd', 'some_list': [1, 2, 4,]}},
        'a1': {'b1': {'c1': 'd1', 'some_list': [10, 20, 40,], 'e': 'whatever'}}}
    thing = splitter(d)
    print(thing['qux'][3].root)

    out = GoDown(condition=(IsDictionary() & HasKey(key='c')))(thing)
    out = list(out)
    print(out)
    print('---')
    has_dict = GoDown(condition=HasDescendant(condition=IsDictionary()))
    out = has_dict.apply_label('HiThere')(thing)
    out = list(out)



import sys
import logging

logging.basicConfig(level=logging.INFO)


class NoisyImportFinder:

    PATH_TRIGGER = '__nanostream__'

    def __init__(self, path_entry):
        self.path_entry = path_entry
        print('Checking {}:'.format(path_entry), end=' ')
        if path_entry != self.PATH_TRIGGER:
            print('wrong finder')
            raise ImportError()
        else:
            print('works')
        return

    def find_module(self, fullname, path=None):
        logging.info(fullname)
        if fullname == self.PATH_TRIGGER:
            logging.info('Loading derived class')
            import pdb
            pdb.set_trace()

        return None

if __name__ == '__main__':
    sys.path_hooks.append(NoisyImportFinder)

    for hook in sys.path_hooks:
        print('Path hook: {}'.format(hook))

    sys.path.insert(0, NoisyImportFinder.PATH_TRIGGER)

    try:
        print('importing target_module')
        import __nanostream__.foo
    except Exception as e:
        print('Import failed:', e)
